#include "config.h"

#include "gb18030.h"

#include <iconv.h>
#include <libintl.h>
#include <string.h>

#include "fcitx/context.h"
#include "fcitx/fcitx.h"
#include "fcitx/keys.h"

FCITX_DEFINE_PLUGIN(fcitx_gb18030, ime, FcitxIMClass) = {createInputMethod,
                                                         NULL};
static ImeState *createImeState() {
    ImeState *state = fcitx_utils_malloc0(sizeof(ImeState));
    state->convert = imeStateConvert;

    return state;
}

// static
void *createInputMethod(FcitxInstance *instance) {
    ImeState *state = createImeState();
    FcitxInstanceRegisterIM(instance, state, "gb18030", _("GB18030"), "gb18030",
                            initInputMethod, NULL, processInput,
                            getCandidateWords, NULL, NULL, NULL, NULL,
                            100 /* make gb18030 place at last */, "zh_CN");
    state->owner = instance;

    return state;
}

// static
boolean initInputMethod(void *arg) {
    ImeState *state = (ImeState *)arg;
    FcitxInstanceSetContext(state->owner, CONTEXT_IM_KEYBOARD_LAYOUT, "us");
    return true;
}

static boolean shouldProcessCharKey(FcitxKeySym sym, unsigned int keyState) {
    return FcitxHotkeyIsHotKeyDigit(sym, keyState) ||
           FcitxHotkeyIsHotKeyLAZ(sym, keyState) ||
           FcitxHotkeyIsHotKeyUAZ(sym, keyState);
}

static boolean isMaxRawInputBufferSizeReached(FcitxInputState *input) {
    return FcitxInputStateGetRawInputBufferSize(input) ==
           getMaxRawInputBufferSize(input);
}

static void insertCharIntoRawInputBuffer(FcitxInputState *input,
                                         FcitxKeySym sym) {
    char *rawInputBuffer = FcitxInputStateGetRawInputBuffer(input);

    rawInputBuffer[FcitxInputStateGetRawInputBufferSize(input)] = sym;
    rawInputBuffer[FcitxInputStateGetRawInputBufferSize(input) + 1] = '\0';
    FcitxInputStateSetRawInputBufferSize(
        input, FcitxInputStateGetRawInputBufferSize(input) + 1);
}

static deleteCharFromRawInputBuffer(FcitxInputState *input) {
    FcitxInputStateSetRawInputBufferSize(
        input, FcitxInputStateGetRawInputBufferSize(input) - 1);

    char *rawInputBuffer = FcitxInputStateGetRawInputBuffer(input);
    rawInputBuffer[FcitxInputStateGetRawInputBufferSize(input)] = '\0';
}

static boolean shouldUpdateCandidateWords(FcitxInputState *input) {
    return FcitxInputStateGetRawInputBufferSize(input) ==
           getMaxRawInputBufferSize(input) - 1;
}

// static
INPUT_RETURN_VALUE
processInput(void *arg, FcitxKeySym sym, unsigned int keyState) {
    ImeState *state = (ImeState *)arg;
    FcitxInputState *input = FcitxInstanceGetInputState(state->owner);

    INPUT_RETURN_VALUE retVal = IRV_TO_PROCESS;

    if (shouldProcessCharKey(sym, keyState)) {
        if (!isMaxRawInputBufferSizeReached(input)) {
            insertCharIntoRawInputBuffer(input, sym);

            if (isMaxRawInputBufferSizeReached(input)) {
                retVal = IRV_TO_PROCESS;
            } else
                retVal = IRV_DISPLAY_CANDWORDS;
        }
    } else if (FcitxHotkeyIsHotKey(sym, keyState, FCITX_BACKSPACE)) {
        if (!FcitxInputStateGetRawInputBufferSize(input))
            return IRV_DONOT_PROCESS_CLEAN;

        deleteCharFromRawInputBuffer(input);

        if (!FcitxInputStateGetRawInputBufferSize(input))
            retVal = IRV_CLEAN;
        else {
            retVal = IRV_DISPLAY_CANDWORDS;
        }
    } else if (FcitxHotkeyIsHotKey(sym, keyState, FCITX_SPACE)) {
        if (!FcitxInputStateGetRawInputBufferSize(input))
            return IRV_TO_PROCESS;

        if (!shouldUpdateCandidateWords(input))
            return IRV_DO_NOTHING;

        retVal = FcitxCandidateWordChooseByIndex(
            FcitxInputStateGetCandidateList(input), 0);
    } else
        return IRV_TO_PROCESS;

    return retVal;
}

// static
FcitxCandidateWord makeCandidateWord(void *owner, char *word) {
    FcitxCandidateWord candidateWord;
    candidateWord.callback = getCandidateWord;
    candidateWord.owner = owner;
    candidateWord.priv = NULL;
    candidateWord.strExtra = NULL;
    candidateWord.strWord = strdup(word);
    candidateWord.wordType = MSG_OTHER;

    return candidateWord;
}

// static
INPUT_RETURN_VALUE getCandidateWords(void *arg) {
    ImeState *state = (ImeState *)arg;
    FcitxInputState *input = FcitxInstanceGetInputState(state->owner);
    const int maxCandidateCount = getMaxCandidateCount(input);

    FcitxCandidateWordList *candidateList =
        FcitxInputStateGetCandidateList(input);
    FcitxCandidateWordSetPageSize(candidateList, maxCandidateCount);
    FcitxCandidateWordSetChoose(candidateList, "0123456789abcdef");

    if (shouldUpdateCandidateWords(input)) {
        initInputCodeBufferFromInput(&(state->inputCodeBuffer), input);

        for (int i = 0; i < maxCandidateCount; i++) {
            state->convert(state);

            FcitxCandidateWord candWord =
                makeCandidateWord(state, state->utf8WordBuffer);

            FcitxCandidateWordAppend(candidateList, &candWord);

            state->inputCodeBuffer.buffer[state->inputCodeBuffer.length - 1]++;
        }
    }
    FcitxInputStateSetCursorPos(input,
                                FcitxInputStateGetRawInputBufferSize(input));
    FcitxMessages *preedit = FcitxInputStateGetPreedit(input);
    FcitxMessagesSetMessageCount(preedit, 0);
    FcitxMessagesAddMessageStringsAtLast(
        preedit, MSG_INPUT, FcitxInputStateGetRawInputBuffer(input));

    return IRV_DISPLAY_CANDWORDS;
}

// 四字节部分采用GB/T11383——1989未采用的0x30～0x39作为对双字节编码扩充的后缀，
// 编码范围为0x81308130～0xFE39FE39
static int is4ByteCode(FcitxInputState *input) {
    char *rawInputBuffer = FcitxInputStateGetRawInputBuffer(input);
    int rawInputBufferSize = FcitxInputStateGetRawInputBufferSize(input);

    return rawInputBufferSize >= 3 && rawInputBuffer[2] == FcitxKey_3;
}

static const size_t MAX_CANDIDATE_COUNT_FOR_TWO_BYTE_CODE = 16;
static const size_t MAX_CANDIDATE_COUNT_FOR_FOUR_BYTE_CODE = 10;
static int getMaxCandidateCount(FcitxInputState *input) {
    return is4ByteCode(input) ? MAX_CANDIDATE_COUNT_FOR_FOUR_BYTE_CODE
                              : MAX_CANDIDATE_COUNT_FOR_TWO_BYTE_CODE;
}

static const size_t TWO_BYTE_CODE_COUNT = 2;
static const size_t FOUR_BYTE_CODE_COUNT = 4;
static int getMaxByteCodeCount(FcitxInputState *input) {
    return is4ByteCode(input) ? FOUR_BYTE_CODE_COUNT : TWO_BYTE_CODE_COUNT;
}

// GB18030使用双字节和四字节对汉字进行编码并且使用16进制数字
// 来表示相应对内码，2个16进制符号编码的数字可以用一个字节可以
// 表示对应对数值
static int getMaxRawInputBufferSize(FcitxInputState *input) {
    return getMaxByteCodeCount(input) * 2;
}

static unsigned char charToValue(char c) {
    if (c >= '0' && c <= '9') {
        return c - '0';
    }

    if (c >= 'a' && c <= 'f') {
        return c - 'a' + 10;
    }

    if (c >= 'A' && c <= 'F') {
        return c - 'A' + 10;
    }

    return 0;
}

static void initInputCodeBuffer(InputCodeBuffer *inputCodeBuffer,
                                size_t inputCodeBufferLength, char *rawBuffer,
                                size_t rawBufferLength) {
    memset(inputCodeBuffer, 0, sizeof(inputCodeBuffer));

    inputCodeBuffer->length = inputCodeBufferLength;

    const int rawInputBufferPastEnd = rawBufferLength;
    for (int i = 0; i < inputCodeBufferLength; i++) {
        int rawInputBufferIndex = i * 2;

        int least4ByteValue = 0;
        if (rawInputBufferIndex + 1 < rawInputBufferPastEnd) {
            least4ByteValue = charToValue(rawBuffer[rawInputBufferIndex + 1]);
        }

        inputCodeBuffer->buffer[i] =
            charToValue(rawBuffer[rawInputBufferIndex]) * 16 + least4ByteValue;
    }
}

// static
void initInputCodeBufferFromInput(InputCodeBuffer *inputCodeBuffer,
                                  FcitxInputState *input) {
    initInputCodeBuffer(inputCodeBuffer, getMaxByteCodeCount(input),
                        FcitxInputStateGetRawInputBuffer(input),
                        getMaxRawInputBufferSize(input));
}

static boolean isInBlacklist(char code[2]) {
    static char blacklist[][2] = {
        {0xfd, 0x9c}, {0xfd, 0x9d}, {0xfd, 0x9e}, {0xfd, 0x9f}, {0xfd, 0xa0},
        {0xfe, 0x40}, {0xfe, 0x41}, {0xfe, 0x47}, {0xfe, 0x49}
    };

    size_t blacklistLength = sizeof(blacklist) / sizeof(blacklist[0]);
    for (size_t i = 0; i < blacklistLength; i++) {
        if (blacklist[i][0] == code[0] && blacklist[i][1] == code[1]) {
            return true;
        }
    }

    return false;
}
// static
void imeStateConvert(ImeState *state) {
    memset(state->utf8WordBuffer, 0, sizeof(state->utf8WordBuffer));
    if (isInBlacklist(state->inputCodeBuffer.buffer)) {
        return;
    }

    IconvStr inputBuffer = state->inputCodeBuffer.buffer;
    size_t inputBufferBytesLeft = state->inputCodeBuffer.length;
    char *outputBuffer = state->utf8WordBuffer;
    size_t outputBufferBytesLeft = sizeof(state->utf8WordBuffer);

    iconv_t cd = iconv_open("utf-8", "gb18030");
    iconv(cd, &inputBuffer, &inputBufferBytesLeft, &outputBuffer,
          &outputBufferBytesLeft);
    iconv_close(cd);

    return state->utf8WordBuffer;
}

// static
INPUT_RETURN_VALUE getCandidateWord(void *arg,
                                    FcitxCandidateWord *candidateWord) {
    ImeState *state = (ImeState *)arg;
    FcitxInputState *input = FcitxInstanceGetInputState(state->owner);

    strcpy(FcitxInputStateGetOutputString(input), candidateWord->strWord);
    return IRV_COMMIT_STRING;
}
